
/**************************************************************************
  stat.pl - Cheng Hu       February - May 2003
  
  Text statistics tools in Prolog for Natural Language Processing.
  
  (1) average word length, standard deviation, skewness, and kurtosis;
  (2) average sentence length, standard deviation, skewness, and kurtosis;
  (3) token count (total numbers of words);
  (4) type count( number of different words);
  (5) type/token ratio;
  (6) vocabulary richness;
  (7) word frequencies;
  (8) first person form percentage;
  (9) text similarity comparison;
  (10) readability index.
***************************************************************************/

% Loading the tokenizer. 

:- ensure_loaded('et.pl').

% tList is a list of tokens, e.g. [w([a,b,c]), w([...]), n(['1','2',..,]),
% s('?')]. sList is a sorted list of tList in alphabetic order, such as 
% [abc, bcd, cbf, ...], and only words are extracted from tList. 
% fTable is a list of words along with frequencies (in negative values)  
% by aphabetic order, like [[-f1, word1], [-f2, word2],...].
% All these three lists are stored in memory by asserta/1.

:- dynamic(tList/1).
:- dynamic(sList/1).
:- dynamic(fTable/1).

% Clearing all dynamic storages.

clear_storage :- retract(tList(_)),
		 retract(sList(_)),
		 retract(fTable(_)).
		 
% store_lists(+L1, +L2, +L3)
% Stores three lists in memory.

store_lists(L1, L2, L3) :- asserta(tList(L1)),
			   asserta(sList(L2)),
			   asserta(fTable(L3)).

% Initializing the three dynamic lists to prevent retract failure when the
% program starts.

:- store_lists(0,0,0).


:- write('Please run query ''?- go'' or ''?- help'' first.\n'). 

go :-   write('To compare the similarity of two text files, \n'), 
	write('two lists of tokens, two lists of atoms, \n'),
	write('or two lists of char lists, query: \n'),
	write('?- file_compare(File1, File2, CosineTheta).\n'),
	write('?- tokens_compare(TokenList1, TokenList2, CosineTheta).\n'),
	write('?- atoms_compare(AtomList1, AtomList2, CosineTheta). Or,\n'),
	write('?- chars_compare(CharList1, CharList2, CosineTheta).\n'),
	write('To run other statistic tools: \n'),
	write('first query ?- stat_file(FileName). \n'),
	write('stat_tokens(TokenList). stat_atoms(AtomList). Or, \n'),
	write('stat_chars(CharList), for the input of a file , \n'),
	write('a token list, an atom list, \n'),
	write('or a list of char lists. Then, for \n'),
	write('word length, query ?- word_length(WL, SD, Skew, Kurtosis).\n'),
	write('sentence length (only for files or token lists), query \n'),
	write('?- sentence_length(SL, SD, Skew, Kurtosis).\n'),
	write('token count, query ?- token_count(TokenCount).\n'),
	write('token type, query ?- type_count(TypeCount).\n'),
	write('type/token ratio, query ?- token_ratio(Ratio).\n'),
	write('vocabulary richness, query ?- richness(R). \n'),
	write('first person form percentage, \n'),
	write('query ?- first_person_percentage(FP).\n'),
	write('readability index (only for files or list of tokens),\n'),
	write('query ?- read_index(Index).\n'),
	write('word frequency table in alphabetic order,\n'),
	write('query ?- word_freqA(WF).\n'),
	write('word frequency table by frequency value,\n'),
	write('query ?- word_freqV(WF).\n'),
	write('frequency of a specific word, query ?- freq(Word, F).\n').

	
% help menu.

help :- go.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% stat_file(+FileName)
% Tokenizes a text file, stores tokens in tList,  stores the sorted list
% in sList and word frequencies in fTable. 

stat_file(FileName) :-  clear_storage,
			tokenize_file(FileName, TokenList),
			token_atom(TokenList, AtomList),
			msort(AtomList, SortedList),
			freq_count(SortedList, FreqTable),
			store_lists(TokenList, SortedList, FreqTable).

% stat_tokens(+TokenList)
% Processes a token list and stores information in tList, sList, and fTable.

stat_tokens(TokenList) :- clear_storage,			  
			  token_atom(TokenList, AtomList),
			  msort(AtomList, SortedList),
			  freq_count(SortedList, FreqTable),
			  store_lists(TokenList, SortedList, FreqTable).

% stat_chars(+CharList)
% Processes a list of plain char lists and stores the information
% in sList and fTable.

stat_chars(CharList) :- clear_storage,			
			charList_atomList(CharList, AtomList),
			msort(AtomList, SortedList),
			freq_count(SortedList, FreqTable),
			store_lists(0, SortedList, FreqTable).

% stat_atoms(+AtomList)
% Processes a list of atoms and stores the information
% in sList and fTable.
			
stat_atoms(AtomList) :- clear_storage,		
			msort(AtomList, SortedList),
			freq_count(SortedList, FreqTable),
			store_lists(0, SortedList, FreqTable).			

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% token_atom(+TokenList, -AtomList)
% From the output of the other routines, extracts just
% the word tokens and converts them to atoms.

token_atom([],[]) :- !.
token_atom([w(Chars)|X], [Atom|Y]) :-   !, atom_chars(Atom, Chars),
					token_atom(X, Y).
					
token_atom([_|X], Y) :- token_atom(X, Y).

% charList_atomList(+CharList, -AtomList)
% Converts a list of char lists to a list of atoms.

charList_atomList([H1|T1], [H2|T2]) :-  atom_chars(H2, H1),
					charList_atomList(T1,T2).
charList_atomList([], []).

% freq_count(+L1, -L2)
% Counts all word frequencies (in negative values, multiplied by -1).

freq_count([H1|T1], [[N,H1]|T2]) :- count_aux(-1, N, H1, T1, TT1),
				    freq_count(TT1, T2).
freq_count([], []).

count_aux(X, Y, Word, [F|R], Rest) :-   Word==F, YY is X-1,
					count_aux(YY, Y, Word, R, Rest), !.
count_aux(X, X, _, R, R) :- !.
count_aux(X, X, _, [], []).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% word_length(-WL, -SD, -SK, -KU)
% Calculates the average word length, standard deviation,
% skewness, and kurtosis. 

word_length(WL, SD, SK, KU) :-  sList(L), wLength_x(L, LengthList), 
				mean_sd(LengthList, WL, SD, SK, KU).			

wLength_x([], []) :- !.
wLength_x([H1|T1], [H2|T2]) :- atom_length(H1, H2), wLength_x(T1, T2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% mean_sd(+NumberList, -Mean, -SD, -SK, -KU)
% Calculates the mean, standard deviatioin, skewness, and kurtosis of 
% a list of numbers.

mean_sd(NumberList, Mean, SD, SK, KU) :- sum(NumberList, Sum),
					 length(NumberList, N),
					 Mean is Sum/N,
					 sd(NumberList, Mean, N, SD),
					 skewness(NumberList, Mean, N, SD,SK), 
					 kurtosis(NumberList, Mean, N, SD,KU).

% sd(+L, +Mean, +N, -SD)
% Calculates the standard deviation of a list of numbers.

sd(_, _, 1, 0) :- !.
sd(L, Mean, N, SD) :-   sd_x(L, Mean, SquaredL),
			sum(SquaredL, NewS), 
			SD is sqrt(NewS/(N-1)).

sd_x([], _, []) :- !.			
sd_x([H1|T1], Mean, [H2|T2]) :- H2 is (H1 - Mean)**2,
				sd_x(T1, Mean, T2).


% Sum of numbers in a list

sum([], 0) :- !.
sum([H|T], Sum) :- sum(T, S), Sum is H + S.

% skewness(+NumberList, +Mean, +N, +SD, -SK)
% Calculates the skewness of a list of numbers.

skewness(_, _, 1, _, 0) :- !.
skewness(_, _, _, 0, 0) :- !.
skewness(L, Mean, N, SD, SK) :- skewness_x(L, Mean, CubeL),
				sum(CubeL, S),
				SK is S/((N-1)*SD*SD*SD).
skewness_x([], _, []) :- !.			
skewness_x([H1|T1], Mean, [H2|T2]) :-   H2 is (H1 - Mean)**3,
					skewness_x(T1, Mean, T2).

% kurtosis(+NumberList, +Mean, +N, +SD, -KU)
% Calculates the kurtosis of a list of numbers.

kurtosis(_, _, 1, _, 0) :- !.
kurtosis(_, _, _, 0, 999999) :- !.	% if the standard deviation is zero,
					% kurtosis is not defined?
kurtosis(L, Mean, N, SD, KU) :- kurtosis_x(L, Mean, NewL),
				sum(NewL, S),
				KU is S/((N-1)*SD*SD*SD*SD)-3.
kurtosis_x([], _, []) :- !.			
kurtosis_x([H1|T1], Mean, [H2|T2]) :-   H2 is (H1 - Mean)**4,
					kurtosis_x(T1, Mean, T2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% sentence_length(-SL, -SD, -SK, -KU)
% Calculates the average sentence length, standard deviation, skewness, and 
% kurtosis. This predicate is only used for files and token lists.
 
sentence_length(SL, SD, SK, KU) :- tList(L), sLength(L, LengthList), 
				   mean_sd(LengthList, SL, SD, SK, KU).

% sLength(+L, -LengthList)
% Counts sentence lengths from a list of tokens.

sLength(L, LengthList) :- setof(X, endofsentence(X, L), L1),
			  sLength_x([0|L1], LengthList).

sLength_x([_], []) :- !.
sLength_x([H1,H2|T], [F|R]) :- F is H2-H1-1, sLength_x([H2|T], R).

% endofsentence(-Pos, +L)
% Finds the position of '.', '?', and '!' in a list of tokens.

endofsentence(Pos, L) :-  nth1(Pos, L, s('.')).
endofsentence(Pos, L) :-  nth1(Pos, L, s('?')).
endofsentence(Pos, L) :-  nth1(Pos, L, s('!')).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% token_count(-TC)
% Counts the total number of words.

token_count(TC) :- sList(L), length(L, TC).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% type_count(-TC)
% Counts the number of different words.

type_count(TC) :- fTable(T), length(T, TC).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% token_ratio(-R)
% Calculates the type/token ratio.

token_ratio(R) :- token_count(TC), type_count(TP), R is TP/TC.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% richness(-R)
% Calculates the richness of vocabulary. 

richness(R) :-  type_count(Type), token_count(Token), 
		R is Type/sqrt(Token).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% word_freqA(-FA)
% Prints word frequency table in alphabetic order and 
% instantiates the argument to the frequency list.

word_freqA(FA) :- write('Word Frequency Table in Alphabet Sequence:\n'),
		  fTable(T), freq_alphabet(T, FA), nl.

% word_freqV(-FV)
% Prints word frequency table by frequency values (high to low) 
% and instantiates the argument to the frequency list.
		
word_freqV(FV) :-  write('Word Frequency Table by Frequency Values:\n'),
		   fTable(T), freq_value(T, FV), nl.	
		
freq_alphabet([[Num,Word]|T1], [[NewNum, Word]|T2]) :-
	NewNum is -1*Num,
	writef('%20l%w', [Word, NewNum]), 
	nl, freq_alphabet(T1, T2).
	
freq_alphabet([], []).

freq_value(T, FV) :- msort(T, ST), freq_alphabet(ST, FV).

% freq(+Word, -Freq)
% Finds the frequency of a word in the text.

freq(Word, Freq) :- fTable(T), member([X, Word], T), Freq is -1*X, !.
freq(_, 0). 	% word not found.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% first_person_percentage(-FP)
% Calculates the percentage of first person forms.

first_person_percentage(FP) :-  fTable(T), first_person_no(T, FN),
				token_count(TC),
				FP is FN*100/TC, !.
				
first_person_percentage(0).	% If there is no first person form,
				% first_person_no/2 will fail.

% first_person_no(+T, -FN)
% Calculates the sum of numbers of all first person forms.

first_person_no(T, FN) :- bagof(X, first_person_x(T, X), Xlist),
			  sum(Xlist, Sum), FN is -1*Sum.
			
first_person_x(T, N) :- first_person_list(FL), member(F, FL),
			member([N, F], T).

% First person forms.
first_person_list([i, me, my, mine, myself, we, us, our, ours, ourselves]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% file_compare(+File1, +File2, -CosTheta) 
% Compares the similarity of two text files.

file_compare(File1, File2, CosTheta) :- stat_file(File1), fTable(Table1),
					stat_file(File2), fTable(Table2),
					comparing(Table1, Table2, CosTheta).

% tokens_compare(+TL1, +TL2, -CosTheta) 
% Compares the similarity of two lists of tokens.

tokens_compare(TL1, TL2, CosTheta) :-   stat_tokens(TL1), fTable(Table1),
					stat_tokens(TL2), fTable(Table2),
					comparing(Table1, Table2, CosTheta).

% atoms_compare(+AL1, +AL2, -CosTheta) 
% Compares the similarity of two lists of atoms.

atoms_compare(AL1, AL2, CosTheta) :-    stat_atoms(AL1), fTable(Table1),
					stat_atoms(AL2), fTable(Table2),
					comparing(Table1, Table2, CosTheta).

% chars_compare(+CL1, +CL2, -CosTheta) 
% Compares the similarity of two lists of plain char lists.

chars_compare(CL1, CL2, CosTheta) :-    stat_chars(CL1), fTable(Table1),
					stat_chars(CL2), fTable(Table2),
					comparing(Table1, Table2, CosTheta).

% compare(+Table1, +Table2, -CosTheta)
% Compares the similarity of two texts through their frequency tables.

comparing(Table1, Table2, CosineTheta) :-  numerator(Table1, Table2, NumList),
					   sum(NumList, Num),
					   denominator(Table1, Table2, Denom),
					   CosineTheta is Num/Denom.
			     
% numerator(+L1, +L2, -L3)
% Calculates the numerator of the cosine(theta) fomular.

numerator([], _, []) :- !.
numerator([[X1,Word]|T1], Table2, [F|R]) :- member([X2, Word], Table2),
					    F is X1*X2,
					    numerator(T1, Table2, R), !.
numerator([_|T], Table2, L) :- numerator(T, Table2, L).					

% demominator(+L1, +L2, -L3)
% Calculates the denominator of the cosine(theta) fomular.

denominator(T1, T2, Denom) :-   square_list(T1, SL1), square_list(T2, SL2), 
				sum(SL1, D1), sum(SL2, D2),
				Denom is sqrt(D1)*sqrt(D2).
				
% square list(+L1, -L2)
% Extracts values from the first list and stores
% their square values in another list.

square_list([], []) :- !.
square_list([[X,_]|T], [F|R]) :- F is X*X, square_list(T, R).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% read_index(-Index)
% Calculates the readability index.
% This predicate is only used for files and lists of tokens.

read_index(Index) :-    tList(L1), syllables(L1, N1),
			token_count(TC),
			sentence_length(AvgSL, _, _, _),
			Percentage is N1*100/TC, 
			Index is 0.4*(AvgSL+Percentage).

% syllables(+TokenList, -N)
% Counts the number of words over 2 syllables long
% in a list of tokens. 

syllables([], 0) :- !.

syllables([w(Chars)|X], N) :-  
	morph(Chars, NewChars),
	length(NewChars, L), L<5, % To have 3 or more syllables,
	syllables(X, N),!.	  % a word must have 5 or more letters.
				
syllables([w(Chars)|X], N) :-   morph(Chars, NewChars),
				syl_x(NewChars, N1),
				N1<3, syllables(X, N),!.
				
syllables([w(_)|X], N) :-  syllables(X, NN), N is NN+1,!.	
				
syllables([_|X], Y) :- syllables(X, Y).

% syl_x(+CharList, -Syllables)
% Finds the number of syllables in a char list.

syl_x([], 0) :- !.
syl_x([C,l,e], 1) :- \+ vowel(C),!.
syl_x([C, e], 0) :- !, \+ vowel(C).
syl_x([H|T], N) :- \+ vowel(H), syl_x(T, N), !.
syl_x([_, H2|T], N) :- vowel(H2), syl_x(T, NN), N is NN+1, !.
syl_x([_|T], N) :- syl_x(T, NN), N is NN+1.


% Simplified morphological analyzer that only removes suffixes which
% contain vowels, such as -es, -ed, -ing.

% morph(?Chars,?Morpheme)
% Giving a char list, it will find its root.

morph([], []) :- !.

morph(Chars, Root) :- split_suffix(Chars, Root).

morph([H|T], [H|T2]) :- morph(T, T2).			

% suffix(?Chars,?Morpheme)
% If Chars is a suffix, Morpheme is the description of it.

suffix([e,d]).
suffix([i,n,g]).


% vowel(?Char)
% Char is a vowel.

vowel(a).  vowel(e).  vowel(i).  vowel(o).  vowel(u).  vowel(y).


% split_suffix(+Characters,-Root)
% Splits a word into root and suffix.
% Fails if there is no suffix.

% Suffixes with doubled consonants in root

split_suffix([V,C,C,V2|Rest],[V,C]) :-
   vowel(V), \+ vowel(C), vowel(V2), suffix([V2|Rest]).

% -es = -s after s (etc.)

split_suffix([s,h,e,s],[s,h]).
split_suffix([c,h,e,s],[c,h]).
split_suffix([s,e,s],[s]).
split_suffix([z,e,s],[z]).
split_suffix([x,e,s],[x]).

split_suffix([F|R], [F]) :- suffix(R).
